@@ -391,14 +391,12 @@ static int _handle_digestmd5_challenge(xmpp_conn_t *conn,
         xmpp_stanza_set_text(authdata, response);
         strophe_free(conn->ctx, response);
 
-        xmpp_stanza_add_child(auth, authdata);
-        xmpp_stanza_release(authdata);
+        xmpp_stanza_add_child_ex(auth, authdata, 0);
 
         handler_add(conn, _handle_digestmd5_rspauth, XMPP_NS_SASL, NULL, NULL,
                     NULL);
 
         send_stanza(conn, auth, XMPP_QUEUE_STROPHE);
-        xmpp_stanza_release(auth);
 
     } else {
         return _handle_sasl_result(conn, stanza, "DIGEST-MD5");
@@ -432,7 +430,6 @@ static int _handle_digestmd5_rspauth(xmpp_conn_t *conn,
         xmpp_stanza_set_name(auth, "response");
         xmpp_stanza_set_ns(auth, XMPP_NS_SASL);
         send_stanza(conn, auth, XMPP_QUEUE_STROPHE);
-        xmpp_stanza_release(auth);
     } else {
         return _handle_sasl_result(conn, stanza, "DIGEST-MD5");
     }
@@ -491,11 +488,9 @@ static int _handle_scram_challenge(xmpp_conn_t *conn,
         xmpp_stanza_set_text(authdata, response);
         strophe_free(conn->ctx, response);
 
-        xmpp_stanza_add_child(auth, authdata);
-        xmpp_stanza_release(authdata);
+        xmpp_stanza_add_child_ex(auth, authdata, 0);
 
         send_stanza(conn, auth, XMPP_QUEUE_STROPHE);
-        xmpp_stanza_release(auth);
 
         rc = 1; /* Keep handler */
     } else {
@@ -625,7 +620,6 @@ static void _auth(xmpp_conn_t *conn)
                     NULL);
 
         send_stanza(conn, auth, XMPP_QUEUE_STROPHE);
-        xmpp_stanza_release(auth);
 
         /* TLS was tried, unset flag */
         conn->tls_support = 0;
@@ -653,7 +647,6 @@ static void _auth(xmpp_conn_t *conn)
                     "ANONYMOUS");
 
         send_stanza(conn, auth, XMPP_QUEUE_STROPHE);
-        xmpp_stanza_release(auth);
 
         /* SASL ANONYMOUS was tried, unset flag */
         conn->sasl_support &= ~SASL_MASK_ANONYMOUS;
@@ -689,14 +682,12 @@ static void _auth(xmpp_conn_t *conn)
         }
         strophe_free(conn->ctx, str);
 
-        xmpp_stanza_add_child(auth, authdata);
-        xmpp_stanza_release(authdata);
+        xmpp_stanza_add_child_ex(auth, authdata, 0);
 
         handler_add(conn, _handle_sasl_result, XMPP_NS_SASL, NULL, NULL,
                     "EXTERNAL");
 
         send_stanza(conn, auth, XMPP_QUEUE_STROPHE);
-        xmpp_stanza_release(auth);
 
         /* SASL EXTERNAL was tried, unset flag */
         conn->sasl_support &= ~SASL_MASK_EXTERNAL;
@@ -754,14 +745,12 @@ static void _auth(xmpp_conn_t *conn)
         }
         xmpp_stanza_set_text(authdata, str);
         strophe_free(conn->ctx, str);
-        xmpp_stanza_add_child(auth, authdata);
-        xmpp_stanza_release(authdata);
+        xmpp_stanza_add_child_ex(auth, authdata, 0);
 
         handler_add(conn, _handle_scram_challenge, XMPP_NS_SASL, NULL, NULL,
                     (void *)scram_ctx);
 
         send_stanza(conn, auth, XMPP_QUEUE_STROPHE);
-        xmpp_stanza_release(auth);
 
         /* SASL SCRAM-SHA-1 was tried, unset flag */
         conn->sasl_support &= ~scram_ctx->alg->mask;
@@ -776,7 +765,6 @@ static void _auth(xmpp_conn_t *conn)
                     NULL);
 
         send_stanza(conn, auth, XMPP_QUEUE_STROPHE);
-        xmpp_stanza_release(auth);
 
         /* SASL DIGEST-MD5 was tried, unset flag */
         conn->sasl_support &= ~SASL_MASK_DIGESTMD5;
@@ -805,14 +793,12 @@ static void _auth(xmpp_conn_t *conn)
         strophe_free(conn->ctx, str);
         strophe_free(conn->ctx, authid);
 
-        xmpp_stanza_add_child(auth, authdata);
-        xmpp_stanza_release(authdata);
+        xmpp_stanza_add_child_ex(auth, authdata, 0);
 
         handler_add(conn, _handle_sasl_result, XMPP_NS_SASL, NULL, NULL,
                     "PLAIN");
 
         send_stanza(conn, auth, XMPP_QUEUE_STROPHE);
-        xmpp_stanza_release(auth);
 
         /* SASL PLAIN was tried */
         conn->sasl_support &= ~SASL_MASK_PLAIN;
@@ -923,19 +909,15 @@ static int _do_bind(xmpp_conn_t *conn, xmpp_stanza_t *bind)
             return 0;
         }
         xmpp_stanza_set_text(text, resource);
-        xmpp_stanza_add_child(res, text);
-        xmpp_stanza_release(text);
-        xmpp_stanza_add_child(bind, res);
-        xmpp_stanza_release(res);
+        xmpp_stanza_add_child_ex(res, text, 0);
+        xmpp_stanza_add_child_ex(bind, res, 0);
         strophe_free(conn->ctx, resource);
     }
 
-    xmpp_stanza_add_child(iq, bind);
-    xmpp_stanza_release(bind);
+    xmpp_stanza_add_child_ex(iq, bind, 0);
 
     /* send bind request */
     send_stanza(conn, iq, XMPP_QUEUE_STROPHE);
-    xmpp_stanza_release(iq);
     return 0;
 }
 
@@ -957,6 +939,13 @@ _handle_features_sasl(xmpp_conn_t *conn, xmpp_stanza_t *stanza, void *userdata)
     if (bind) {
         ns = xmpp_stanza_get_ns(bind);
         conn->bind_required = ns != NULL && strcmp(ns, XMPP_NS_BIND) == 0;
+        bind = xmpp_stanza_copy(bind);
+        if (!bind) {
+            disconnect_mem_error(conn);
+            return 0;
+        }
+    } else {
+        conn->bind_required = 0;
     }
 
     /* check whether session establishment is required */
@@ -974,12 +963,6 @@ _handle_features_sasl(xmpp_conn_t *conn, xmpp_stanza_t *stanza, void *userdata)
         conn->sm_state->sm_support = 1;
     }
 
-    bind = xmpp_stanza_copy(bind);
-    if (!bind) {
-        disconnect_mem_error(conn);
-        return 0;
-    }
-
     /* we are expecting either <bind/> and <session/> since this is a
        XMPP style connection or we <resume/> the previous session */
 
@@ -999,7 +982,6 @@ _handle_features_sasl(xmpp_conn_t *conn, xmpp_stanza_t *stanza, void *userdata)
         strophe_snprintf(h, sizeof(h), "%u", conn->sm_state->sm_handled_nr);
         xmpp_stanza_set_attribute(resume, "h", h);
         send_stanza(conn, resume, XMPP_QUEUE_SM_STROPHE);
-        xmpp_stanza_release(resume);
         handler_add(conn, _handle_sm, XMPP_NS_SM, NULL, NULL, NULL);
     }
     /* if bind is required, go ahead and start it */
@@ -1008,7 +990,9 @@ _handle_features_sasl(xmpp_conn_t *conn, xmpp_stanza_t *stanza, void *userdata)
         _do_bind(conn, bind);
     } else {
         /* can't bind, disconnect */
-        xmpp_stanza_release(bind);
+        if (bind) {
+            xmpp_stanza_release(bind);
+        }
         strophe_error(conn->ctx, "xmpp",
                       "Stream features does not allow "
                       "resource bind.");
@@ -1080,12 +1064,10 @@ _handle_bind(xmpp_conn_t *conn, xmpp_stanza_t *stanza, void *userdata)
             xmpp_stanza_set_name(session, "session");
             xmpp_stanza_set_ns(session, XMPP_NS_SESSION);
 
-            xmpp_stanza_add_child(iq, session);
-            xmpp_stanza_release(session);
+            xmpp_stanza_add_child_ex(iq, session, 0);
 
             /* send session establishment request */
             send_stanza(conn, iq, XMPP_QUEUE_STROPHE);
-            xmpp_stanza_release(iq);
         }
 
         if (conn->sm_state->sm_support && !conn->sm_disable) {
@@ -1100,7 +1082,6 @@ _handle_bind(xmpp_conn_t *conn, xmpp_stanza_t *stanza, void *userdata)
                 xmpp_stanza_set_attribute(enable, "resume", "true");
             handler_add(conn, _handle_sm, XMPP_NS_SM, NULL, NULL, NULL);
             send_stanza(conn, enable, XMPP_QUEUE_SM_STROPHE);
-            xmpp_stanza_release(enable);
         }
 
         if (!conn->session_required) {
@@ -1359,15 +1340,13 @@ static void _auth_legacy(xmpp_conn_t *conn)
         goto err_free;
     xmpp_stanza_set_name(query, "query");
     xmpp_stanza_set_ns(query, XMPP_NS_AUTH);
-    xmpp_stanza_add_child(iq, query);
-    xmpp_stanza_release(query);
+    xmpp_stanza_add_child_ex(iq, query, 0);
 
     child = xmpp_stanza_new(conn->ctx);
     if (!child)
         goto err_free;
     xmpp_stanza_set_name(child, "username");
-    xmpp_stanza_add_child(query, child);
-    xmpp_stanza_release(child);
+    xmpp_stanza_add_child_ex(query, child, 0);
 
     authdata = xmpp_stanza_new(conn->ctx);
     if (!authdata)
@@ -1379,29 +1358,25 @@ static void _auth_legacy(xmpp_conn_t *conn)
     }
     xmpp_stanza_set_text(authdata, str);
     strophe_free(conn->ctx, str);
-    xmpp_stanza_add_child(child, authdata);
-    xmpp_stanza_release(authdata);
+    xmpp_stanza_add_child_ex(child, authdata, 0);
 
     child = xmpp_stanza_new(conn->ctx);
     if (!child)
         goto err_free;
     xmpp_stanza_set_name(child, "password");
-    xmpp_stanza_add_child(query, child);
-    xmpp_stanza_release(child);
+    xmpp_stanza_add_child_ex(query, child, 0);
 
     authdata = xmpp_stanza_new(conn->ctx);
     if (!authdata)
         goto err_free;
     xmpp_stanza_set_text(authdata, conn->pass);
-    xmpp_stanza_add_child(child, authdata);
-    xmpp_stanza_release(authdata);
+    xmpp_stanza_add_child_ex(child, authdata, 0);
 
     child = xmpp_stanza_new(conn->ctx);
     if (!child)
         goto err_free;
     xmpp_stanza_set_name(child, "resource");
-    xmpp_stanza_add_child(query, child);
-    xmpp_stanza_release(child);
+    xmpp_stanza_add_child_ex(query, child, 0);
 
     authdata = xmpp_stanza_new(conn->ctx);
     if (!authdata)
@@ -1418,14 +1393,12 @@ static void _auth_legacy(xmpp_conn_t *conn)
         xmpp_disconnect(conn);
         return;
     }
-    xmpp_stanza_add_child(child, authdata);
-    xmpp_stanza_release(authdata);
+    xmpp_stanza_add_child_ex(child, authdata, 0);
 
     handler_add_id(conn, _handle_legacy, "_xmpp_auth1", NULL);
     handler_add_timed(conn, _handle_missing_legacy, LEGACY_TIMEOUT, NULL);
 
     send_stanza(conn, iq, XMPP_QUEUE_STROPHE);
-    xmpp_stanza_release(iq);
     return;
 
 err_free:
